"""
A module to hold the models for the zum library regarding the requests.
"""

from typing import Any, Dict


class Request:

    """
    Class to encapsulate the request structure.
    """

    def __init__(
        self,
        route: str,
        method: str,
        params: Dict[str, Any],
        body: Dict[str, Any],
    ) -> None:
        self._route = route
        self.method = method
        self.params = params
        self.body = body

    @property
    def route(self) -> str:
        """Interpolate params into the route."""
        return self._route.format(**self.params)



"""
Module for the core configuration logic of zum.
"""

from pathlib import Path
from typing import Any, Dict

import tomlkit

from zum.configs.errors import InvalidConfigFileError, MissingConfigFileError


def search_for_config_file(file_name: str) -> None:
    """Searches for the config file. If it doesn't exist, raises an error."""
    if not Path(file_name).is_file():
        raise MissingConfigFileError(f"Missing config file '{file_name}'.")


def retrieve_config_file(file_name: str) -> Dict[str, Any]:
    """Opens the config file and returns the parsed dictionary."""
    with open(file_name, "r") as raw_configs:
        configs = raw_configs.read()
    return tomlkit.parse(configs)


def validate_configs(configs: Dict[str, Any]) -> None:
    """Validate that the configs are not malformed."""
    validate_metadata(configs)
    validate_endpoints(configs)


def validate_metadata(configs: Dict[str, Any]) -> None:
    """Validate that the metadata key of the configs are not malformed."""
    if "metadata" not in configs:
        raise InvalidConfigFileError("Missing 'metadata' section of the config file")
    if not isinstance(configs["metadata"], dict):
        raise InvalidConfigFileError(
            "The 'metadata' section shold be a dictionary or mapping."
        )
    if "server" not in configs["metadata"]:
        raise InvalidConfigFileError(
            "Missing 'server' value from the 'metadata' section of the config file"
        )


def validate_endpoints(configs: Dict[str, Any]) -> None:
    """Validate that the endpoints key of the configs are not malformed."""
    if "endpoints" not in configs:
        raise InvalidConfigFileError("Missing 'endpoints' section of the config file")
    if not isinstance(configs["endpoints"], dict):
        raise InvalidConfigFileError(
            "The 'endpoints' section shold be a dictionary or mapping."
        )
    if len(configs["endpoints"].keys()) == 0:
        raise InvalidConfigFileError(
            "At least one endpoint is required on the config file"
        )



"""
A module to route the zum CLI traffic.
"""

from argparse import ArgumentParser
from typing import Any, List, Optional

import zum
from zum.engine import Engine


def dispatcher(*args: Any, **kwargs: Any) -> None:
    """
    Main CLI method, recieves the command line action and dispatches it to
    the corresponding method.
    """
    engine = Engine()

    parser = generate_parser(engine.actions)
    parsed_args = parser.parse_args(*args, **kwargs)

    engine.execute(parsed_args.action[0], parsed_args.params)
    log(engine.output)


def generate_parser(actions_list: List[str]) -> ArgumentParser:
    """Generates the action parser."""
    warning = "Beware! No config file was found!" if not actions_list else ""

    # Create parser
    parser = ArgumentParser(
        description="Command line interface tool for zum.", epilog=warning
    )

    # Add version command
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"zum version {zum.__version__}",
    )

    if actions_list:
        # Add action argument
        parser.add_argument("action", choices=actions_list, nargs=1)

        # Add params
        parser.add_argument("params", nargs="*")

    return parser


def log(data: Optional[str]) -> None:
    """Logs a string to the console."""
    print(data)



"""
The main module for the zum library.
"""

import json
from typing import Any, List, Optional

import httpx

from zum.configs.core import (
    retrieve_config_file,
    search_for_config_file,
    validate_configs,
)
from zum.configs.errors import InvalidConfigFileError, MissingConfigFileError
from zum.constants import CONFIG_FILE_NAME
from zum.executor import execute
from zum.requests.core import generate_request
from zum.requests.validations import validate_raw_endpoint


class Engine:

    """
    Class to handle the overall behaviour of zum.
    """

    def __init__(self) -> None:
        self.__output: Optional[str] = None
        self.__exception: Optional[Exception] = None

        try:
            search_for_config_file(CONFIG_FILE_NAME)
            configs = retrieve_config_file(CONFIG_FILE_NAME)
            validate_configs(configs)
        except (MissingConfigFileError, InvalidConfigFileError) as error:
            configs = {}
            self.__exception = error

        self.__metadata = configs.get("metadata", {})
        self.__endpoints = configs.get("endpoints", {})

    @property
    def output(self) -> Optional[str]:
        """Returns the internal state of the output."""
        return self.__output

    @property
    def actions(self) -> List[str]:
        """Returns a list with the possible actions."""
        return list(self.__endpoints.keys())

    def execute(self, instruction: str, arguments: List[Any]) -> None:
        """
        Executes the main zum logic.
        """
        self.__validate_configurations()
        validate_raw_endpoint(self.__endpoints[instruction])
        request = generate_request(self.__endpoints[instruction], arguments)
        response = execute(self.__metadata["server"], request)
        self.__handle_response(response)

    def __validate_configurations(self) -> None:
        """Validates that the configurations were correctly loaded."""
        if self.__exception:
            raise self.__exception

    def __handle_response(self, response: httpx.Response) -> None:
        """Handles the response."""
        try:
            self.__output = json.dumps(
                response.json(), indent=2, sort_keys=False, ensure_ascii=False
            )
        except TypeError:
            self.__output = response.text



"""
Module for the requests validations of zum.
"""

from typing import Any, Dict

from zum.constants import HTTP_METHODS
from zum.requests.errors import InvalidEndpointDefinitionError


def validate_raw_endpoint(raw_endpoint: Dict[str, Any]) -> None:
    """Validates that the endpoint's dictionary data is valid."""
    validate_raw_endpoint_route(raw_endpoint)
    validate_raw_endpoint_method(raw_endpoint)


def validate_raw_endpoint_route(raw_endpoint: Dict[str, Any]) -> None:
    """Validates that the endpoint's dictionary route value is valid."""
    if "route" not in raw_endpoint:
        raise InvalidEndpointDefinitionError(
            "Missing 'route' attribute for the endpoint"
        )
    route = raw_endpoint["route"]
    if not isinstance(route, str):
        raise InvalidEndpointDefinitionError(
            "The 'route' attribute of the endpoint must be a string"
        )


def validate_raw_endpoint_method(raw_endpoint: Dict[str, Any]) -> None:
    """Validates that the endpoint's dictionary method value is valid."""
    if "method" not in raw_endpoint:
        raise InvalidEndpointDefinitionError(
            "Missing 'method' attribute for the endpoint"
        )
    method = raw_endpoint["method"]
    if not isinstance(method, str):
        raise InvalidEndpointDefinitionError(
            "The 'method' attribute for the endpoint must be a string"
        )
    if method.strip().lower() not in HTTP_METHODS:
        raise InvalidEndpointDefinitionError(
            "Invalid 'method' value for the endpoint (not a valid HTTP method)"
        )



"""
A module to hold the custom exceptions for the zum library regarding the
configuration file.
"""


class MissingConfigFileError(Exception):
    """An exception for when the config file is missing."""


class InvalidConfigFileError(Exception):
    """An exception for when the config file is invalid."""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zum', 'zum.configs', 'zum.requests']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.16.1,<0.17.0', 'tomlkit>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['zum = zum.cli:dispatcher']}

setup_kwargs = {
    'name': 'zum',
    'version': '0.0.3',
    'description': 'Stop writing scripts to test your APIs. Call them as CLIs instead.',
    'long_description': '<p align="center">\n    <a href="https://github.com/daleal/zum">\n        <img src="https://zum.daleal.dev/assets/zum-250x250.png">\n    </a>\n</p>\n\n<h1 align="center">Zum</h1>\n\n<p align="center">\n    <em>\n        Stop writing scripts to test your APIs. Call them as CLIs instead.\n    </em>\n</p>\n\n<p align="center">\n<a href="https://pypi.org/project/zum" target="_blank">\n    <img src="https://img.shields.io/pypi/v/zum?label=version&logo=python&logoColor=%23fff&color=306998" alt="PyPI - Version">\n</a>\n\n<a href="https://github.com/daleal/zum/actions?query=workflow%3Atests" target="_blank">\n    <img src="https://img.shields.io/github/workflow/status/daleal/zum/tests?label=tests&logo=python&logoColor=%23fff" alt="Tests">\n</a>\n\n<a href="https://codecov.io/gh/daleal/zum" target="_blank">\n    <img src="https://img.shields.io/codecov/c/gh/daleal/zum?label=coverage&logo=codecov&logoColor=ffffff" alt="Coverage">\n</a>\n\n<a href="https://github.com/daleal/zum/actions?query=workflow%3Alinters" target="_blank">\n    <img src="https://img.shields.io/github/workflow/status/daleal/zum/linters?label=linters&logo=github" alt="Linters">\n</a>\n</p>\n\n**Zum** (German word roughly meaning "_to the_" or "_to_" depending on the context, pronounced `/tsʊm/`) is a tool that lets you describe an API using a [TOML](https://toml.io/en/) file and then call that API using your command line. This means that **the days of writing custom scripts to help you develop each of your APIs are over**. Just create a `zum.toml`, describe your API and forget about maintaining more code!\n\n## Installing\n\nInstall using pip!\n\n```sh\npip install zum\n```\n\n## Usage\n\n### The config file\n\nThe first thing that you need to do in order to use `zum` is to describe the API that you\'re trying to ping using a config file, named `zum.toml`. This TOML file should contain the following keys:\n\n#### `metadata`\n\nThe `metadata` key contains everything that is not data about an endpoint of the API itself, but that is needed in order to query the API. This key should contain the following values:\n\n- `server`: The base URL where the API is hosted.\n\nAs an example, the first lines of your `zum.toml` file for a development environment should probably look similar to this:\n\n```toml\n[metadata]\nserver = "http://localhost:8000"\n```\n\nThis indicates to `zum` that the API endpoints are located at `http://localhost:8000`. Easy enough, right?\n\n#### `endpoints`\n\nThe `endpoints` key contains every endpoint that you want to be able to access from `zum`. Each endpoint should also have a `route` value, a `method` value and may include a `params` value and a `body` value. Let\'s see an example:\n\n```toml\n[endpoints.my-endpoint-name]\nroute = "/endpoint-name"\nmethod = "post"\n```\n\nNotice that the header of the section consists of `endpoints.{something}`. **That `{something}` will be the name of your endpoint**. That means that, on the example, to query the endpoint, all you need to do is to run:\n\n```sh\nzum my-endpoint-name\n```\n\nWith the existing configuration, `zum` will make a `POST` HTTP request to `http://localhost:8000/endpoint-name`. Just 5 lines on a TOML file!\n\nThe endpoint configuration will be discussed more on a [dedicated section](#endpoints).\n\n### Endpoints\n\nUp until now, the examples shown are **really** simple. Rarely does an API endpoint not include some parameters on the URL or some request body. The idea of `zum` is to keep everything extremely simple, so let\'s see how to use URL parameters and the request body.\n\n#### Parameters\n\nImagine that you have an API with an endpoint `/entity/:id` that returns an entity with the id `:id`. You would probably like to be able to just `zum entity 3` to get the entity with `:id` corresponding to `3`. Well, that\'s **exactly** what `zum` does. Let\'s first define the endpoint:\n\n```toml\n[endpoints.my-entity]\nroute = "/entity/{id}"\nmethod = "get"\nparams = ["id"]\n```\n\nThis configuration tells `zum` that it should replace the route string `{id}` for whatever `id` it receives from the command line as an argument (the command for that configuration should be `zum my-entity 3`). That makes sense. But why is the `params` value an array? Well, let\'s imagine that you want to search for all the appearances of some string on the entity model. The API endpoint will probably receive a `?query=` parameter. So let\'s describe this new endpoint:\n\n```toml\n[endpoints.search]\nroute = "/entity/{id}?query={query}"\nmethod = "get"\nparams = ["id", "query"]\n```\n\nNow, you can run something like:\n\n```sh\nzum search 3 mystring\n```\n\nThis will search `"mystring"` on the entity with id `3`. **But order matters**. Let\'s imagine that to you, it makes more sense to write the query string first. Then, your definition should be:\n\n```toml\n[endpoints.search]\nroute = "/entity/{id}?query={query}"\nmethod = "get"\nparams = ["query", "id"]\n```\n\nNow, you can run something like:\n\n```sh\nzum search mystring 3\n```\n\nThe query will be exactly the same. **This means that the array tells `zum` in which order to interpret the CLI parameters**.\n\n#### Request body\n\nJust like the parameters, the request body gets defined as an array:\n\n```toml\n[endpoints.create-entity]\nroute = "/entity"\nmethod = "post"\nbody = ["name", "city"]\n```\n\nTo run this endpoint, you just need to run:\n\n```sh\nzum create-entity dani Santiago\n```\n\nThis will send a `POST` request to `http://localhost:8000/entity` with the following request body:\n\n```json\n{\n    "name": "dani",\n    "city": "Santiago"\n}\n```\n\nAs always, order matters.\n\n```toml\n[endpoints.create-entity]\nroute = "/entity"\nmethod = "post"\nbody = ["city", "name"]\n```\n\nNow, to get the same result as before, you should run:\n\n```sh\nzum create-entity Santiago dani\n```\n\n#### What about both?\n\nOf course, sometimes you need to use both parameters **and** request bodies. For example, if you wanted to create a nested entity, you would need to use the parent\'s id as a parameter and the new entity data as a request body. Let\'s describe this situation!\n\n```toml\n[endpoints.create-nested]\nroute = "/entity/{id}"\nmethod = "post"\nparams = ["id"]\nbody = ["name", "city"]\n```\n\nNow, you can call the endpoint using:\n\n```sh\nzum create-nested 69 dani Santiago\n```\n\nThis will call `POST /entity/69` with the following request body:\n\n```json\n{\n    "name": "dani",\n    "city": "Santiago"\n}\n```\n\nAs you can probably tell, `zum` receives the `params` first on the CLI, and then the `body`. In _pythonic_ terms, what `zum` does is kind of _unpacks_ both arrays consecutively, something like the following:\n\n```py\narguments = [*params, *body]\nzum(arguments)\n```\n\n## `zum.toml` example\n\nHere\'s a simple `zum.toml` file example:\n\n```toml\n[metadata]\nserver = "http://localhost:8000"\n\n[endpoints.my-entity]\nroute = "/entity/{id}"\nmethod = "get"\nparams = ["id"]\n\n[endpoints.search]\nroute = "/entity/{id}?query={query}"\nmethod = "get"\nparams = ["id", "query"]\n\n[endpoints.create-entity]\nroute = "/entity"\nmethod = "post"\nbody = ["name", "city"]\n\n[endpoints.create-nested]\nroute = "/entity/{id}"\nmethod = "post"\nparams = ["id"]\nbody = ["name", "city"]\n```\n\nWith that config file (using a hypothetical existing API), you could `GET /entity/420` to get the entity with id `420`, `GET /entity/420?query=nice` to search for the appearances of the word `nice` on the model of the entity with id `420`, `POST /entity` with some request body to create a new entity and `POST /entity/69` with some request body to create a new nested entity, child of the entity with id `69`.\n\n## Developing\n\nClone the repository:\n\n```sh\ngit clone https://github.com/daleal/zum.git\n\ncd zum\n```\n\nRecreate environment:\n\n```sh\nmake get-poetry\nmake venv-with-dependencies\n```\n\nRun the linters:\n\n```sh\nmake black flake8 isort mypy pylint\n```\n\nRun the tests:\n\n```sh\nmake tests\n```\n\n## Resources\n\n- [Issue Tracker](https://github.com/daleal/zum/issues/)\n',
    'author': 'Daniel Leal',
    'author_email': 'dlleal@uc.cl',
    'maintainer': 'Daniel Leal',
    'maintainer_email': 'dlleal@uc.cl',
    'url': 'https://github.com/daleal/zum',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
Module for the core requests logic of zum.
"""

from typing import Any, Dict, List, Optional, Tuple

from zum.requests.errors import MissingEndpointParamsError
from zum.requests.models import Request


def generate_request(raw_endpoint: Dict[str, Any], arguments: List[str]) -> Request:
    """
    Given the raw endpoint data and the arguments list, generates a Request
    object with the data necessary to execute said request.
    """
    params, remaining_arguments = reduce_arguments(
        raw_endpoint.get("params"), arguments
    )
    body, _ = reduce_arguments(raw_endpoint.get("body"), remaining_arguments)
    return Request(raw_endpoint["route"], raw_endpoint["method"], params, body)


def reduce_arguments(
    keys: Optional[List[str]], arguments: List[Any]
) -> Tuple[Dict[str, Any], List[Any]]:
    """
    Given a :keys array of strings, maps the first :keys.length elements
    of the :arguments array to the :keys elements as keys, returning the
    mapped elements and the rest of the :arguments array.
    """
    if keys is None:
        return {}, arguments
    if len(arguments) < len(keys):
        raise MissingEndpointParamsError(
            "Invalid amount of arguments passed to the command."
        )
    return (
        {x[0]: x[1] for x in zip(keys, arguments)},
        arguments[len(keys) :],
    )
